/*******************************************************************************
* © 2011-2017, Cypress Semiconductor Corporation
* or a subsidiary of Cypress Semiconductor Corporation. All rights
* reserved.
* 
* This software, including source code, documentation and related
* materials (“Software”), is owned by Cypress Semiconductor
* Corporation or one of its subsidiaries (“Cypress”) and is protected by
* and subject to worldwide patent protection (United States and foreign),
* United States copyright laws and international treaty provisions.
* Therefore, you may use this Software only as provided in the license
* agreement accompanying the software package from which you
* obtained this Software (“EULA”).
* 
* If no EULA app_objectlies, Cypress hereby grants you a personal, non-
* exclusive, non-transferable license to copy, modify, and compile the
* Software source code solely for use in connection with Cypress’s
* integrated circuit products. Any reproduction, modification, translation,
* compilation, or representation of this Software except as specified
* above is prohibited without the express written permission of Cypress.
* 
* Disclaimer: THIS SOFTWARE IS PROVIDED AS-IS, WITH NO
* WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING,
* BUT NOT LIMITED TO, NONINFRINGEMENT, IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE. Cypress reserves the right to make
* changes to the Software without notice. Cypress does not assume any
* liability arising out of the app_objectlication or use of the Software or any
* product or circuit described in the Software. Cypress does not
* authorize its products for use in any products where a malfunction or
* failure of the Cypress product may reasonably be expected to result in
* significant property damage, injury or death (“High Risk Product”). By
* including Cypress’s product in a High Risk Product, the manufacturer
* of such system or app_objectlication assumes all risk of such use and in doing
* so agrees to indemnify Cypress against all liability.
********************************************************************************/
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using PP_COM_Wrapper;
using System.Threading;

namespace C_Sharp
{
    class Program
    {

        const int NUMBER_OF_PSOCS = 2;
        private static object AcquireLockingVar = new object();
        private static object ProgrammingLockingVar = new object();
        static PSoCProgrammerCOM_ObjectClass pp_test = new PSoCProgrammerCOM_ObjectClass();
        static PSoCProgrammerCOM_ObjectClass[] pp = new PSoCProgrammerCOM_ObjectClass[NUMBER_OF_PSOCS];

        static string m_sLastError;
        static int threadCounter = 0, delayAdder = 0;

        //Distinguishing identifier of PSoC3/5 families
        const int LEOPARD_ID = 0xE0;
        const int PANTHER_ID = 0xE1;

        static void PPCOM_INIT()
        {

            for (int i = 0; i < NUMBER_OF_PSOCS; i++)
            {
                pp[i] = new PSoCProgrammerCOM_ObjectClass();
            }
        }

        static bool SUCCEEDED(int hr)
        {
            return hr >= 0;
        }

        //Check JTAG ID of device - identify family PSoC3 or PSoC5
        static int GetGenerationByJtagID(byte[] JtagID)
        {
            int distinguisher = (((JtagID[0] & 0x0F) << 4) | (JtagID[1] >> 4));
            return distinguisher;
        }

        static bool IsPSoC3ES3(byte[] jtagID)
        {
            if (GetGenerationByJtagID(jtagID) == LEOPARD_ID)
            {
                if ((jtagID[0] >> 4) >= 0x01) return true; //silicon version==0x01 saved in bits [4..7]
            }                                            //For ES0-2==0x00, ES3==0x01
            return false;
        }

        static bool Is_PSoC5_TM_ID(byte[] jtagID)
        {
            //0E0xx069
            return ((jtagID[0] & 0xFF) == 0x0E) && ((jtagID[1] & 0xF0) == 0x10) && (jtagID[3] == 0x69);
        }

        static bool Is_PSoC5_LP_ID(byte[] id)
        {
            //Check whether SWD ID belongs to PSoC5LP
            //PSoC5LP: 0x1BA01477 (SWD/JTAG read ID read request retuns CM3 ID always)
            //         0x2E1xx069 (LP) - this Device ID must be read from PANTHER_DEVICE_ID reg (0x4008001C)
            //2E0xx069
            return ((id[0] & 0xFF) >= 0x2E) && ((id[1] & 0xF0) == 0x10) && (id[3] == 0x69);
        }

        static int ClosePort(PSoCProgrammerCOM_ObjectClass pp_object)
        {
            //Close port when all operations are complete, so it will be available for other app_objectlications
            int hr;
            string strError;

            hr = pp_object.ClosePort(out strError);

            return hr;
        }

        static int CheckHexAndDeviceCompatibility(PSoCProgrammerCOM_ObjectClass pp_object, out bool result)
        {
            int hr;
            object jtagID;
            byte[] hexJtagID, chipJtagID;
            result = false;

            hr = pp_object.DAP_GetJtagID(out jtagID, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            chipJtagID = jtagID as byte[];

            hr = pp_object.HEX_ReadJtagID(out jtagID, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            hexJtagID = jtagID as byte[];

            result = true;
            for (byte i = 0; i < 4; i++)
            {
                if (hexJtagID[i] != chipJtagID[i])
                {
                    result = false;
                    break;
                }
            }
            return 0;
        }

        static int GetTotalFlashRowsCount(int flashSize, out int totalRows)
        {
            int rowSize = 256;
            totalRows = flashSize / rowSize;
            return 0;
        }

        static int CheckSum_All(int flashSize, out uint cs, PSoCProgrammerCOM_ObjectClass pp_object)
        {
            int hr = 0;
            cs = 0;

            object arrays;
            hr = pp_object.PSoC3_GetSonosArrays(enumSonosArrays.ARRAY_FLASH, out arrays, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;

            int RowsToChecksum;
            GetTotalFlashRowsCount(flashSize, out RowsToChecksum);

            //Checksum all the Flash arrays
            Array arrayInfo = arrays as Array;
            for (int i = arrayInfo.GetLowerBound(1); i <= arrayInfo.GetUpperBound(1); i++)
            {
                int arrayID = (int)arrayInfo.GetValue(0, i);
                int arraySize = (int)arrayInfo.GetValue(1, i);
                //Get info about flash array
                int rowSize, rowsPerArray, eccPresence;
                hr = pp_object.PSoC3_GetFlashArrayInfo(arrayID, out rowSize, out rowsPerArray, out eccPresence, out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                //Find number of rows in array to be checksumed
                if ((RowsToChecksum - rowsPerArray) < 0) rowsPerArray = RowsToChecksum;
                //Calculate checksum of the array
                uint arrayChecksum = 0;
                hr = pp_object.PSoC3_CheckSum(arrayID, 0, rowsPerArray, out arrayChecksum, out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                //Sum checksum
                cs += arrayChecksum;
                //Update number of rows to be checksumed
                RowsToChecksum -= rowsPerArray;
                if (RowsToChecksum <= 0) break;
            }

            return hr;
        }

        static int IsNvlUpdatePossible(PSoCProgrammerCOM_ObjectClass pp_object, out bool state)
        {
            int hr, power, voltage;
            hr = pp_object.GetPower1(out power, out voltage, out m_sLastError);
            if (!SUCCEEDED(hr)) { state = false; return hr; }
            if (voltage > 3700)
                state = false;
            else
                state = true;
            return hr;
        }

        static int ProgramNvlArrays(PSoCProgrammerCOM_ObjectClass pp_object,enumSonosArrays nvlArrayType)
        {
            int hr;
            object arrays;
            hr = pp_object.PSoC3_GetSonosArrays(nvlArrayType, out arrays, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            Array arrayInfo = arrays as Array;
            int addrHex = 0;
            for (int i = arrayInfo.GetLowerBound(1); i <= arrayInfo.GetUpperBound(1); i++)
            {
                int arrayID = (int)arrayInfo.GetValue(0, i);
                int arraySize = (int)arrayInfo.GetValue(1, i);
                //Read data from Hex file
                object data;
                if (nvlArrayType == enumSonosArrays.ARRAY_NVL_USER)
                    hr = pp_object.HEX_ReadNvlCustom(addrHex, arraySize, out data, out m_sLastError);
                else //enumSonosArrays.ARRAY_NVL_WO_LATCHES
                    hr = pp_object.HEX_ReadNvlWo(addrHex, arraySize, out data, out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                byte[] hexData = data as byte[];
                addrHex += arraySize;
                //Read data from device
                hr = pp_object.PSoC3_ReadNvlArray(arrayID, out data, out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                byte[] chipData = data as byte[];
                //Compare data from Chip against corresponding Hex-block
                if (chipData.Length != hexData.Length)
                {
                    m_sLastError = "Hex file's NVL array differs from corresponding device's one!";
                    return -1;
                }
                bool fIdentical = true;
                for (int j = 0; j < arraySize; j++)
                {
                    if (chipData[j] != hexData[j])
                    {
                        fIdentical = false;
                        break;
                    }
                }
                if (fIdentical) continue; //Arrays are equal, skip programming, goto following array
                //Verify voltage range for TM silicon. Do not allow TM programming if voltage is greater than 3.3 V
                bool state;
                object id;
                hr = pp_object.DAP_GetJtagID(out id, out m_sLastError);
                if (Is_PSoC5_TM_ID(id as byte[]))
                {
                    IsNvlUpdatePossible(pp_object, out state);
                    if (!state)
                    {
                        m_sLastError = ((nvlArrayType == enumSonosArrays.ARRAY_NVL_USER) ? "User NVLs" : "WO NVLs") +
                                       " update failed. This operation can be completed in voltage range 1.8 - 3.3 V only.";
                        return -1;
                    }
                }
                //Program NVL array
                hr = pp_object.PSoC3_WriteNvlArray(arrayID, hexData, out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                //PSoC3 ES3 supp_objectort - check whether ECCEnable bit is changed and reacquire device
                if (nvlArrayType == enumSonosArrays.ARRAY_NVL_USER)
                {
                    object JtagID;
                    hr = pp_object.DAP_GetJtagID(out JtagID, out m_sLastError);
                    if (!SUCCEEDED(hr)) return hr;
                    //ES3 and probably newer revisions
                    if (IsPSoC3ES3(JtagID as byte[]) || Is_PSoC5_LP_ID(JtagID as byte[]))
                    {
                        bool eccEnableChanged = false;
                        if (hexData.Length >= 4)
                            eccEnableChanged = ((hexData[3] ^ chipData[3]) & 0x08) == 0x08;
                        //need to reacquire chip if EccEnable bit was changed to app_objectly it for flash rows.
                        if (eccEnableChanged)
                            hr = pp_object.DAP_Acquire(out m_sLastError);
                    }
                    return hr;
                }
            }
            return hr;
        }

        static int GetEccOption(PSoCProgrammerCOM_ObjectClass pp_object, int arrayID)
        {
            int hr;
            int rowSize, rowsPerArray, eccPresence;
            hr = pp_object.PSoC3_GetFlashArrayInfo(arrayID, out rowSize, out rowsPerArray, out eccPresence, out m_sLastError);
            int eccHwStatus;
            hr = pp_object.PSoC3_GetEccStatus(out eccHwStatus, out m_sLastError); //get ecc status of the acquired device
            if (!SUCCEEDED(hr)) return hr;
            int eccOption = ((eccPresence != 0) && (eccHwStatus == 0)) ? 1 : 0; //take into account data from the Config area
            return eccOption;
        }

        static int ProgramFlashArrays(PSoCProgrammerCOM_ObjectClass pp_object, int flashSize)
        {
            int hr;
            object arrays;
            hr = pp_object.PSoC3_GetSonosArrays(enumSonosArrays.ARRAY_FLASH, out arrays, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;

            int flashProgrammed = 0;
            //Program Flash arrays
            Array arrayInfo = arrays as Array;
            for (int i = arrayInfo.GetLowerBound(1); i <= arrayInfo.GetUpperBound(1); i++)
            {
                int arrayID = (int)arrayInfo.GetValue(0, i);
                int arraySize = (int)arrayInfo.GetValue(1, i);
                //Program flash array from hex file taking into account ECC (Config Data)
                int rowSize, rowsPerArray, eccPresence;
                hr = pp_object.PSoC3_GetFlashArrayInfo(arrayID, out rowSize, out rowsPerArray, out eccPresence, out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                int eccOption = GetEccOption(pp_object, arrayID); //take into account data from the Config area
                for (int rowID = 0; rowID < rowsPerArray; rowID++)
                {
                    hr = pp_object.PSoC3_ProgramRowFromHex(arrayID, rowID, eccOption, out m_sLastError);
                    if (!SUCCEEDED(hr)) return hr;
                    //Limit programming to the device flash size
                    flashProgrammed += rowSize;
                    if (flashProgrammed >= flashSize) return hr;
                }
            }
            return hr;
        }

        static int Allow_Reset_Hard(PSoCProgrammerCOM_ObjectClass pp_object)
        {
            //------------------------------------------------------------------------
            //CDT#130393, post #31, 33
            //Need to set "allow_rst_hard", so toggling XRES works as Power Cycle.
            //If that bit is not set no all subsystems of PSoC5 LP are reset.

            object JtagID;
            int hr = pp_object.DAP_GetJtagID(out JtagID, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            //App_objectly to PSoC5 LP only
            if (Is_PSoC5_LP_ID(JtagID as byte[]))
            {
                string strError;
                //Set 'allow_rst_hrd' bit in MLOGIC_DEBUG register (see PSoC5 IROS: 001-43078, Book 2)
                hr = pp_object.DAP_WriteIO(0x400046E8, 0x00000040, out strError);
            }
            return hr;
        }

        static int ProgramAll(PSoCProgrammerCOM_ObjectClass pp_object)
        {
            if (pp_object == null) return -1;

            string path = Path.GetDirectoryName(Environment.CurrentDirectory);
            string filePath = path + "\\LP035_HelloWorld_Blinky.hex";
            int hr;
            //Port Initialization
            //Setup Power - "2.5V" and internal
            hr = pp_object.SetPowerVoltage("5.0", out m_sLastError);
            if (!SUCCEEDED(hr))
            {
                Console.Write("Set power voltage failed!");
                return hr;
            }
            hr = pp_object.PowerOn(out m_sLastError);
            if (!SUCCEEDED(hr))
            {
                Console.Write("Power ON failed!");
                return hr;
            }
            //Set protocol, connector and frequency
            hr = pp_object.SetProtocol(enumInterfaces.SWD, out m_sLastError); //SWD-protocol
            if (!SUCCEEDED(hr))
            {
                Console.Write("Set protocol failed!");
                return hr;
            }
            hr = pp_object.SetProtocolConnector(1, out m_sLastError); //10-pin connector
            if (!SUCCEEDED(hr))
            {
                Console.Write("Connector failed!");
                return hr;
            };
            hr = pp_object.SetProtocolClock(enumFrequencies.FREQ_01_6, out m_sLastError); //1.6 MHz clock on SWD bus
            // Set Hex File
            int hexImageSize;
            hr = pp_object.HEX_ReadFile(filePath, out hexImageSize, out m_sLastError);
            if (!SUCCEEDED(hr))
            {
                Console.Write("Read hex failed!");
                return hr;
            };
            //Set Acquire Mode
            pp_object.SetAcquireMode("Reset", out m_sLastError);

            //Random random = new Random();
            //int delayAdded = random.Next(1, 8);

            //Thread.Sleep(15000 * delayAdded);

            lock (ProgrammingLockingVar)
            {
                //The "Programming Flow" proper
                //Acquire Device
                hr = pp_object.DAP_Acquire(out m_sLastError);
                if (!SUCCEEDED(hr))
                {
                    Console.Write("DAP Acquire failed!");
                    return hr;
                };
                //Check Hex File and Device compatibility
                bool fCompatibility;
                hr = CheckHexAndDeviceCompatibility(pp_object, out fCompatibility);
                if (!SUCCEEDED(hr)) return hr;
                if (!fCompatibility)
                {
                    m_sLastError = "The Hex file does not match the acquired device, please connect the app_objectropriate device";
                    return -1;
                }
                //Erase All
                hr = pp_object.PSoC3_EraseAll(out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
                //Program User NVL arrays
                hr = ProgramNvlArrays(pp_object, enumSonosArrays.ARRAY_NVL_USER);
                if (!SUCCEEDED(hr)) return hr;
                //Program Flash arrays
                hr = ProgramFlashArrays(pp_object, hexImageSize);
                //Protect All arrays
                hr = pp_object.PSoC3_ProtectAll(out m_sLastError);
                if (!SUCCEEDED(hr)) return hr;
            }

            //CheckSum
            short hexChecksum;
            uint flashChecksum;
            hr = CheckSum_All(hexImageSize, out flashChecksum, pp_object);
            if (!SUCCEEDED(hr)) return hr;
            hr = pp_object.HEX_ReadChecksum(out hexChecksum, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            flashChecksum &= 0xFFFF;
            int hexChecks = hexChecksum & 0xFFFF;

            if (flashChecksum != hexChecks)
            {
                m_sLastError = "Mismatch of Checksum: Expected 0x" + flashChecksum.ToString("X") + ", Got 0x" + hexChecksum.ToString("X");
                return -1;
            }
            else
                Console.WriteLine("Checksum 0x" + flashChecksum.ToString("X"));

            //Release PSoC3 device
            Allow_Reset_Hard(pp_object); //needed for PSoC5 LP only
            hr = pp_object.DAP_ReleaseChip(out m_sLastError);

            return hr;
        }

        static private int UpgradeBlock(PSoCProgrammerCOM_ObjectClass pp_object)
        {
            if (pp_object == null) return -1;

            int hr;
            //Port Initialization
            //Setup Power - "2.5V" and internal
            hr = pp_object.SetPowerVoltage("2.5", out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            hr = pp_object.PowerOn(out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            //Set protocol, connector and frequency
            hr = pp_object.SetProtocol(enumInterfaces.SWD, out m_sLastError); //SWD-protocol
            if (!SUCCEEDED(hr)) return hr;
            hr = pp_object.SetProtocolConnector(1, out m_sLastError); //10-pin connector
            if (!SUCCEEDED(hr)) return hr;
            hr = pp_object.SetProtocolClock(enumFrequencies.FREQ_03_0, out m_sLastError); //<=3.0 MHz for Read operations

            //Set Acquire Mode
            pp_object.SetAcquireMode("Reset", out m_sLastError);

            //Acquire Device
            hr = pp_object.DAP_Acquire(out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            //Write Row, use PSoC3_WriteRow() instead PSoC3_ProgramRow()
            byte[] writeData = new byte[288]; //User and Config area of the Row (256+32)
            for (int i = 0; i < 288; i++) writeData[i] = (byte)i;
            int arrayID = 0, rowID = 255;
            hr = pp_object.PSoC3_WriteRow(arrayID, rowID, writeData, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            //Verify Row - only user area (without Config one)
            object readData;
            hr = pp_object.PSoC3_ReadRow(arrayID, rowID, 0, out readData, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            byte[] readRow = readData as byte[];
            for (int i = 0; i < readRow.Length; i++) //check 256 bytes
            {
                if (readRow[i] != writeData[i])
                {
                    hr = -1;
                    break;
                }
            }
            if (!SUCCEEDED(hr))
            {
                m_sLastError = "Verification of User area failed!";
                return hr;
            }
            //Verify Row - Config are only
            hr = pp_object.PSoC3_ReadRow(arrayID, rowID, 1, out readData, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;
            readRow = readData as byte[];
            for (int i = 0; i < readRow.Length; i++) //check 32 bytes
            {
                if (readRow[i] != writeData[256 + i])
                {
                    hr = -1;
                    break;
                }
            }
            if (!SUCCEEDED(hr))
            {
                m_sLastError = "Verification of Config area failed!";
                return hr;
            }

            //Release PSoC3 chip
            Allow_Reset_Hard(pp_object); //needed for PSoC5 LP only
            hr = pp_object.DAP_ReleaseChip(out m_sLastError);

            return hr;
        }

        #region DDKA 

        static private int getNumberOfPorts()
        {

            int numberOfPortsConnected, hr;

            object portArray;

            hr = pp_test.GetPorts(out portArray, out m_sLastError);
            if (!SUCCEEDED(hr)) return hr;

            string[] ports = portArray as string[];
            if (ports.Length <= 0)
            {
                m_sLastError = "Connect any Programmer to PC";
                return -1;
            }

            numberOfPortsConnected = ports.Length;

            return numberOfPortsConnected;
        }

        static private string[] getValidPortNames()
        {

            int hr, validPorts = 0;
            object portArray;

            hr = pp_test.GetPorts(out portArray, out m_sLastError);

            string[] ports = portArray as string[];

            for (int i = 0; i < ports.Length; i++)
            {
                if (ports[i].StartsWith("MiniProg") || ports[i].StartsWith("TrueTouchBridge") || ports[i].StartsWith("DVKProg"))
                {
                    validPorts++;
                }
            }

            string[] portName = new string[validPorts];

            for (int i = 0; i < ports.Length; i++)
            {
                if (ports[i].StartsWith("MiniProg3") || ports[i].StartsWith("TrueTouchBridge") || ports[i].StartsWith("DVKProg"))
                {
                    portName[i] = ports[i];
                }
            }

            return portName;

        }


        static int ProgramPSoC(PSoCProgrammerCOM_ObjectClass pp_object, string portName)
        {

            int hr;

            hr = pp_object.OpenPort(portName, out m_sLastError);

            threadCounter++;

            Console.Write("Running Thread #" + threadCounter + " \r\n");

            //Execute Programming of the silicon
            hr = ProgramAll(pp_object); //hr = UpgradeBlock();

            //Close Port in the end, so port is available for other app_objectlications
            ClosePort(pp_object);

            return hr;
        }

        static private int gangProgram(string[] portNames)
        {

            int hr;

            if (pp == null) return -1;

            int numberOfValidPorts = portNames.Length;
            Console.Write("The number of valid ports connnected are: " + portNames.Length + "\r\n");

            Thread[] threadsArray = new Thread[numberOfValidPorts];

            for (int j = 0; j < portNames.Length; j++)
            {
                int localNum = j;
                threadsArray[j] = new Thread(() => ProgramPSoC(pp[localNum], portNames[localNum]));
                Console.Write("Thread with port name: " + portNames[localNum] + " created\r\n");
            }

            for (int i = 0; i < portNames.Length; i++)
            {
                threadsArray[i].Start();
            }

            for (int i = 0; i < portNames.Length; i++)
            {
                threadsArray[i].Join();
            }

            return 0;
        }
        #endregion


        static void Main(string[] args)
        {
            Console.WriteLine("Program All using COM-object interface only\r\n");
            int hr;
            string str;

            PPCOM_INIT();

            int numberOfPortsConnected = getNumberOfPorts();
            if (numberOfPortsConnected == 0) Console.WriteLine("No ports connected\r\n");

            string[] portNames = getValidPortNames();
            //Console.Write("The number of valid ports connnected are: " + portNames.Length + "\r\n");

            hr = gangProgram(portNames);

            //hr = Execute();

            if (SUCCEEDED(hr)) str = "Succeeded!";
            else str = "Failed! " + m_sLastError;
            Console.WriteLine(str);

            Console.ReadKey();
        }
    }
}
